package evidence

import (
	"encoding/json"
	"errors"
	"regexp"
	"strings"
	"unicode/utf8"

	"git.huawei.com/evidence/common"
	"git.huawei.com/evidence/identity"
	"git.huawei.com/evidence/model"

	contractapi "git.huawei.com/poissonsearch/wienerchain/contract/docker-container/contract-go/contractapi"
)

var typeMap = make(map[string]string)

func init() {
	typeMap["hash"] = "hash"
}

const (
	MaxIterNum               = 200
	MaxProjectIdLength       = 64
	MaxFileNameLength        = 255
	MaxUserNameLength        = 10
	MaxIdentityTypeLength    = 20
	MaxCopyRighterTypeLength = 20
	MaxIdentityIdLength      = 40
	MaxContentLength         = 1024

	HashLength = 64
	EvIdLength = 32

	ChargeType = "normal"

	sIdPattern  = "^[a-zA-Z0-9]{1,40}"
	bIdPattern  = "^[a-zA-Z0-9_]{1,80}"
	hashPattern = "^[a-zA-Z0-9]{64}"
	evIdPattern = "^[a-zA-Z0-9]{32}"

	HashAlreadyExistErrCode   = "10000101 "
	EvIdAlreadyExistErrCode   = "10000102 "
	SIdBIdAlreadyExistErrCode = "10000103 "

	InputParamInvalidErrCode = "10000003 "
)

func LoadEvidenceApi(funcMap map[string]common.DCMAPI) {
	// evidence
	funcMap["createEvidence"] = CreateEvidence
	funcMap["queryEvidenceByHash"] = QueryEvidenceByHash
	funcMap["queryEvidenceByEvID"] = QueryEvidenceByEvID

	funcMap["queryEvidenceBySIdAndBId"] = QueryEvidenceBySIdAndBId
	funcMap["queryEvidenceBySId"] = QueryEvidenceBySId
}

var CreateEvidence = func(stub contractapi.ContractStub, args [][]byte) ([]byte, error) {
	if len(args) != 1 {
		return nil, errors.New("the params number is not 1")
	}
	evidence := &model.Evidence{}
	if err := json.Unmarshal(args[0], &evidence); err != nil {
		return nil, err
	}
	sender, err := getAddress(stub)
	if err != nil {
		return nil, errors.New("get addr fail,please use correct cert")
	}
	if err = verifyParams(evidence); err != nil {
		return nil, err
	}
	evidence.CreatorAddr = sender.String()
	if err = verifyBusinessParam(stub, evidence, sender); err != nil {
		return nil, err
	}
	err = saveEvidence(stub, evidence, sender)
	return nil, err
}

var QueryEvidenceByHash = func(stub contractapi.ContractStub, args [][]byte) ([]byte, error) {
	hash := string(args[0])

	if !matchString(hashPattern, hash) {
		return nil, errors.New("hash param invalid")
	}
	hashKey, err := common.CreateCompositeKey(common.EvidencePrefix, []string{hash})
	if err != nil {
		return nil, err
	}
	value, err := stub.GetKV(hashKey)
	if err != nil {
		return nil, err
	}
	key, err := common.CreateCompositeKey(common.EvidencePrefix, []string{string(value)})
	if err != nil {
		return nil, err
	}
	return stub.GetKV(key)
}
var QueryEvidenceByEvID = func(stub contractapi.ContractStub, args [][]byte) ([]byte, error) {
	evID := string(args[0])
	if !matchString(evIdPattern, evID) {
		return nil, errors.New("evID param invalid")
	}
	key, err := common.CreateCompositeKey(common.EvidencePrefix, []string{evID})
	value, err := stub.GetKV(key)
	if err != nil && len(value) > 0 {
		return nil, err
	}
	return value, nil
}
var QueryEvidenceBySIdAndBId = func(stub contractapi.ContractStub, args [][]byte) ([]byte, error) {
	if len(args) != 2 {
		return nil, errors.New("args num is not 2")
	}
	sId := string(args[0])
	bId := string(args[1])
	err := checkSIdAndBId(sId, bId)
	if err != nil {
		return nil, err
	}
	sender, err := getAddress(stub)
	if err != nil {
		return nil, errors.New("get addr fail,please use correct cert")
	}
	sIdBIdKey, err := common.CreateCompositeKey(common.EvidencePrefix, []string{sender.String(), sId, bId})
	if err != nil {
		return nil, err
	}
	value, err := stub.GetKV(sIdBIdKey)
	if err != nil {
		return nil, err
	}
	key, err := common.CreateCompositeKey(common.EvidencePrefix, []string{string(value)})
	if err != nil {
		return nil, err
	}
	return stub.GetKV(key)
}
var QueryEvidenceBySId = func(stub contractapi.ContractStub, args [][]byte) ([]byte, error) {
	sId := string(args[0])
	if !matchString(sIdPattern, sId) {
		return nil, errors.New("s_id param invalid")
	}
	sender, err := getAddress(stub)
	if err != nil {
		return nil, errors.New("get addr fail,please use correct cert")
	}
	iter, err := GetIteratorByCompositeKey(stub, common.EvidencePrefix, []string{sender.String(), sId})
	if err != nil {
		return nil, errors.New("GetIteratorByCompositeKey err " + err.Error())
	}
	b := make([]string, 0)
	var i uint32
	for iter.Next() {
		b = append(b, iter.Key())
		i++
		if i == MaxIterNum {
			break
		}
	}

	iter.Close()
	result, err := json.Marshal(&b)
	if err != nil {
		return nil, err
	}
	return result, nil
}

func matchString(pattern, str string) bool {
	compile, err := regexp.Compile(pattern)
	if err != nil {
		return false
	}
	return compile.FindString(str) == str
}

func getAddress(stub contractapi.ContractStub) (identity.Address, error) {
	creatorBytes, err := stub.GetPemCert()
	if err != nil {
		return identity.ZeroAddress, errors.New(err.Error() + " failed to get creator")
	}

	callerAddr, err := identity.CertToAddr(creatorBytes)
	if err != nil {
		return identity.ZeroAddress, errors.New(err.Error() + " fail to convert cert to address")
	}
	return identity.ConvAddress(callerAddr)
}
func GetIteratorByCompositeKey(stub contractapi.ContractStub, prefix string, attrs []string) (contractapi.Iterator, error) {
	key, err := common.CreateCompositeKey(prefix, attrs)
	if err != nil {
		return nil, err
	}
	return stub.GetIterator(key, key+string(common.MaxUnicodeValue))
}

func verifyParams(evidence *model.Evidence) error {
	_, ok := typeMap[evidence.TypeData]
	if !ok {
		return errors.New(InputParamInvalidErrCode + "type param invalid")
	}
	if len(evidence.Hash) != HashLength {
		return errors.New(InputParamInvalidErrCode + "hash param invalid")
	}
	if len(evidence.EvID) != EvIdLength {
		return errors.New(InputParamInvalidErrCode + "ev_id param invalid")
	}
	if evidence.ProjectId == "" || len(evidence.ProjectId) > MaxProjectIdLength {
		return errors.New(InputParamInvalidErrCode + "project_id param invalid")
	}
	if evidence.FileName == "" || utf8.RuneCountInString(evidence.FileName) > MaxFileNameLength {
		return errors.New(InputParamInvalidErrCode + "file_name param invalid")
	}
	if len(evidence.UserName) > MaxUserNameLength {
		return errors.New(InputParamInvalidErrCode + "user_name param invalid")
	}
	if len(evidence.IdentityType) > MaxIdentityTypeLength {
		return errors.New(InputParamInvalidErrCode + "identity_type param invalid")
	}
	if len(evidence.CopyRighterType) > MaxCopyRighterTypeLength {
		return errors.New(InputParamInvalidErrCode + "copy_righter_type param invalid")
	}
	if len(evidence.IdentityId) > MaxIdentityIdLength {
		return errors.New(InputParamInvalidErrCode + "copy_righter_type param invalid")
	}
	if evidence.ChargeType != ChargeType {
		return errors.New(InputParamInvalidErrCode + "charge_type param invalid")
	}
	if len([]byte(evidence.Content)) > MaxContentLength {
		return errors.New(InputParamInvalidErrCode + "content param invalid")
	}
	return nil
}

func hashIsExist(stub contractapi.ContractStub, evidence *model.Evidence) error {
	hashKey, err := common.CreateCompositeKey(common.EvidencePrefix, []string{evidence.Hash})
	if err != nil {
		return err
	}
	value, err := stub.GetKV(hashKey)
	if err != nil {
		return err
	}
	if len(value) == 0 {
		return nil
	}
	return errors.New(HashAlreadyExistErrCode + "the hash already exist")
}

func checkSIdAndBId(sId, bId string) error {
	if !matchString(sIdPattern, sId) {
		return errors.New("s_id param invalid")
	}
	if !matchString(bIdPattern, bId) {
		return errors.New("b_id param invalid")
	}
	return nil
}

func sIdAndBIdIsExist(stub contractapi.ContractStub, senderAddr string, evidence *model.Evidence) error {

	sIdBIdKey, err := common.CreateCompositeKey(common.EvidencePrefix, []string{senderAddr, evidence.SId, evidence.BId})
	if err != nil {
		return err
	}
	value, err := stub.GetKV(sIdBIdKey)
	if err != nil {
		return err
	}
	if len(value) == 0 {
		return nil
	}
	return errors.New(SIdBIdAlreadyExistErrCode + "s_id and b_id  composite key already exist")
}

func saveEvidence(stub contractapi.ContractStub, evidence *model.Evidence, sender identity.Address) error {

	evIDKey, err := common.CreateCompositeKey(common.EvidencePrefix, []string{evidence.EvID})
	if err != nil {
		return errors.New("ev_id CreateCompositeKey fail " + err.Error())
	}
	marshal, err := json.Marshal(evidence)
	if err != nil {
		return err
	}
	if err = stub.PutKV(evIDKey, marshal); err != nil {
		return err
	}
	hashKey, err := common.CreateCompositeKey(common.EvidencePrefix, []string{evidence.Hash})
	if err != nil {
		return err
	}
	if err = stub.PutKV(hashKey, []byte(evidence.EvID)); err != nil {
		return err
	}
	if sIdAnBIdEmpty(evidence) {
		return nil
	}
	sidBidKey, err := common.CreateCompositeKey(common.EvidencePrefix, []string{sender.String(), evidence.SId, evidence.BId})
	if err != nil {
		return err
	}
	if err = stub.PutKV(sidBidKey, []byte(evidence.EvID)); err != nil {
		return err
	}
	return nil
}

func sIdAnBIdEmpty(evidence *model.Evidence) bool {
	return evidence.SId == "" && evidence.BId == ""
}
func sIdAnBIdExist(evidence *model.Evidence) bool {
	return evidence.SId != "" && evidence.BId != ""
}
func verifyBusinessParam(stub contractapi.ContractStub, evidence *model.Evidence, sender identity.Address) error {
	if !(sIdAnBIdEmpty(evidence) || sIdAnBIdExist(evidence)) {
		return errors.New("sid and bid both need to exist")
	}
	evidence.Hash = strings.ToLower(evidence.Hash)
	err := hashIsExist(stub, evidence)
	if err != nil {
		return err
	}
	evIDKey, err := common.CreateCompositeKey(common.EvidencePrefix, []string{evidence.EvID})
	if err != nil {
		return errors.New(" ev_id CreateCompositeKey fail " + err.Error())
	}
	value, err := stub.GetKV(evIDKey)
	if err == nil && len(value) > 0 {
		return errors.New(EvIdAlreadyExistErrCode + "the ev_id already exist")
	}
	if sIdAnBIdEmpty(evidence) {
		return nil
	}
	err = checkSIdAndBId(evidence.SId, evidence.BId)
	if err != nil {
		return err
	}
	err = sIdAndBIdIsExist(stub, sender.String(), evidence)
	if err != nil {
		return err
	}
	return nil
}
