package identity

import (
	"bytes"
	"crypto/sha256"
	"crypto/x509"
	"encoding/hex"
	"encoding/pem"
	"errors"
	"fmt"

	"git.huawei.com/evidence/common"
)

// Address defines the user's identifier in ledger
type Address [common.AddressSize]byte

var ZeroAddress = Address{}

// IsZero returns true if address is invalid zero address.
func (a Address) IsZero() bool {
	bit := byte(0)
	for _, b := range a {
		bit |= b
	}
	return bit == 0
}

func (a Address) IsValid() bool {
	return !a.IsZero()
}

func (a Address) String() string {
	return hex.EncodeToString(a[:])
}

func (a Address) IDType() string {
	return common.IDTypeAddress
}

func (a Address) EmptyIdentity() interface{} {
	return ZeroAddress
}

// EqualsAddrStr returns true if address string converted same address.
func (a Address) EqualsStr(addr string) bool {
	return a.String() == addr
}

// Bytes converts address to bytes.
func (a Address) Bytes() []byte {
	return a[:]
}

// Equals returns true if address is same.
func (a Address) Equals(addr interface{}) bool {
	addrT, ok := addr.(Address)
	if !ok {
		return false
	}
	return bytes.Equal(a[:], addrT[:])
}

// ConvAddress from address bytes to address type.
func ConvAddress(bs []byte) (Address, error) {
	if len(bs) != common.AddressSize {
		return ZeroAddress, fmt.Errorf("address must be 20 bytes, but %d bytes", len(bs))
	}
	var addr Address
	copy(addr[:], bs)
	return addr, nil
}

func ConvStr2Address(addr string) (Address, error) {
	addrBytes, err := hex.DecodeString(addr)
	if err != nil {
		return ZeroAddress, errors.New("invalid address format")
	}
	return ConvAddress(addrBytes)
}

// CertToAddr transfer user's cert to address
func CertToAddr(creator []byte) ([]byte, error) {
	bl, _ := pem.Decode(creator)
	if bl == nil {
		return nil, errors.New("no pem data found")
	}

	cert, err := x509.ParseCertificate(bl.Bytes)
	if err != nil {
		return nil, errors.New(err.Error() + " failed to parse certificate")
	}

	pubkeyBytes, err := x509.MarshalPKIXPublicKey(cert.PublicKey)
	if err != nil {
		return nil, errors.New(err.Error() + " unable to marshal public key")
	}

	// We want the last 160 bits of the sha3-256 sum as the address
	sum := sha256.Sum256(pubkeyBytes)
	return sum[common.Len256AddrToLen160Addr:], nil
}

func GetCertOrg(creator []byte) (string, error) {
	bl, _ := pem.Decode(creator)
	if bl == nil {
		return "", errors.New("no pem data found")
	}

	cert, err := x509.ParseCertificate(bl.Bytes)
	if err != nil {
		return "", errors.New(err.Error() + " failed to parse certificate")
	}
	if len(cert.Subject.Organization) == 0 || len(cert.Subject.Organization[0]) != common.AddressLen {
		return "", errors.New("user cert org error")
	}
	return cert.Subject.Organization[0], nil
}
