package identity

import (
	"encoding/json"
	"errors"
	"regexp"
	"strings"

	"git.huawei.com/evidence/common"
	contractapi "git.huawei.com/poissonsearch/wienerchain/contract/docker-container/contract-go/contractapi"
)

type PublicKeyType string

type Attribute struct {
	// eg last_name
	Name string `json:"name"`
	// eg int, string, fileurl...
	Type string `json:"type"`

	Description string `json:"description"`
}

type CredentialApplySchema struct {
	Type string `json:"type"`
	// eg Huawei Enterprise Certification
	Name        string `json:"name"`
	Description string `json:"description"`
	// the applyer need provider Attributes to apply the verifiable credential
	Attributes []Attribute `json:"attributes"`
}

type Service struct {
	Id    string  `json:"id"`
	Type  string  `json:"type"`
	Price float64 `json:"price,omitempty"`
	// this is a url, service description information needs to be published on the official website
	ServiceEndpoint string `json:"serviceEndpoint"`
	// Option. Indicate which public key is responsible for this service. if empty, default is master key
	Controller string `json:"controller,omitempty"`
	// option.
	CredentialApplySchema CredentialApplySchema `json:"credentialApplySchema,omitempty"`
}

type RecoveryOperateMessage struct {
	DID       string       `json:"did"`
	Publickey DocPublicKey `json:"publicKey"`
}

type UpdateServiceMessage struct {
	DID      string    `json:"did"`
	Services []Service `json:"services"`
}

type DocPublicKey struct {
	// public key id
	Id         string        `json:"id"`
	Type       PublicKeyType `json:"type"`
	Controller string        `json:"controller"`
	// For different useage
	PublicKeyPem    string `json:"publicKeyPem,omitempty"`
	PublicKeyHex    string `json:"publicKeyHex,omitempty"`
	PublicKeyBase58 string `json:"publicKeyBase58,omitempty"`
}

type Proof struct {
	Creator        string `json:"creator"`
	Type           string `json:"type"`
	Created        string `json:"created"`
	SignatureValue string `json:"signatureValue"`
}

type Document struct {
	// Currently fixed, Uniform  https://www.w3.org/ns/did/v1
	Context string `json:"@context"`
	// did, eg: did:hwid:21tDAKCERh95uGgKbJNHYp
	Id string `json:"id"`
	// public key list
	PublicKey []DocPublicKey `json:"publicKey"`
	// the controllers' public id of this Did
	Authentication []string `json:"authentication"`
	// Option. ues to change Authentication
	Recovery string `json:"recovery"`
	//Option
	Service []Service `json:"service,omitempty"`
	Proof   Proof     `json:"proof,omitempty"`
	Created string    `json:"created"`
	Updated string    `json:"updated"`
	Status  string    `json:"status"`
	//Prevent replay attacks default 1
	//Version         int64        `json:"version"`
}

type DID string

func (d DID) Equals(did interface{}) bool {
	didT, ok := did.(DID)
	if !ok {
		return false
	}
	return didT == d
}
func (d DID) EqualsStr(str string) bool {
	return str == d.String()
}
func (d DID) IsValid() bool {
	return !d.EqualsStr("")
}
func (d DID) String() string {
	return string(d)
}

func (d DID) IDType() string {
	return common.IDTypeDID
}
func (d DID) EmptyIdentity() interface{} {
	var dt DID
	return dt
}

func ConvStr2DID(str string) (DID, error) {
	if !IsDID(str) {
		return "", errors.New("invalid did string")
	}
	return DID(str), nil
}

func IsDID(did string) bool {
	if match, err := regexp.MatchString(common.DIDRegx, did); err != nil || !match {
		return false
	}
	return true
}

func (d DID) CheckAuthority(stub contractapi.ContractStub, creator []byte) error {
	docByte, err := QueryDID(stub, [][]byte{[]byte(d.String())})
	if err != nil {
		return err
	}
	document := &Document{}
	if err := json.Unmarshal(docByte, document); err != nil {
		return errors.New("parse did doc error" + err.Error())
	}
	masterKey, err := getDocumentPublicKey(document, d.String())
	if err != nil {
		return err
	}
	if masterKey.PublicKeyPem != string(creator) {
		return errors.New("no authority")
	}
	return nil
}

func LoadDIDAPI(funcMap map[string]common.DCMAPI) {
	funcMap["registerDID"] = RegisterDID
	funcMap["queryDID"] = QueryDID
	funcMap["queryAllDID"] = QueryAllDID
}

var RegisterDID = func(stub contractapi.ContractStub, args [][]byte) ([]byte, error) {
	if len(args) != 1 {
		return nil, errors.New("need document json string")
	}
	documentStr := args[0]
	document := &Document{}

	if err := json.Unmarshal(documentStr, document); err != nil {
		return nil, errors.New("doc format error")
	}
	// 1. check did
	if !IsDID(document.Id) {
		return nil, errors.New("did format error")
	}
	if document.PublicKey == nil || document.Authentication == nil {
		return nil, errors.New("doc error publicKey is empty")
	}
	if len(document.PublicKey) > 10 {
		return nil, errors.New("number of publicKey should less than 10")
	}

	docBytes, err := stub.GetKV(document.Id)
	if err != nil {
		return nil, errors.New("did get error " + err.Error())
	}
	if docBytes != nil {
		return nil, errors.New("did(" + document.Id + ") already exist")
	}
	certBytes, err := stub.GetPemCert()
	if err != nil {
		return nil, errors.New("get cert info error")
	}
	// check authority
	isProofPass := false
	for _, auth := range document.Authentication {
		for _, pk := range document.PublicKey {
			if auth == pk.Id {
				if string(certBytes) == pk.PublicKeyPem {
					isProofPass = true
					break
				}
			}
		}
		if isProofPass {
			break
		}
	}
	if !isProofPass {
		return nil, errors.New("the request creator not the public key user")
	}
	// check service
	if document.Service != nil {
		for _, v := range document.Service {
			if v.Type == "" || v.ServiceEndpoint == "" {
				return nil, errors.New("service type and service endpoint is empty")
			}
		}
	}
	document.Status = common.DocumentStatusActive
	document.Context = common.DIDContextVersion
	docBytes, err = json.Marshal(document)
	if err != nil {
		return nil, errors.New("document marshal error: " + err.Error())
	}
	if err := stub.PutKV(document.Id, docBytes); err != nil {
		return nil, errors.New("document store error: " + err.Error())
	}
	if err := stub.SaveComIndex(common.IDTypeDID, []string{}, document.Id); err != nil {
		return nil, errors.New("save index error: " + err.Error())
	}
	return docBytes, nil
}

var QueryDID = func(stub contractapi.ContractStub, args [][]byte) ([]byte, error) {
	if len(args) != 1 {
		return nil, errors.New("need did string")
	}
	piece := strings.Split(string(args[0]), common.DIDFragmentTag)
	did := piece[0]

	docBytes, err := stub.GetKV(did)
	if err != nil {
		return nil, errors.New("did get error: " + err.Error())
	}
	if docBytes == nil {
		return nil, errors.New("did (" + did + ") not found")
	}
	return docBytes, nil
}

var QueryAllDID = func(stub contractapi.ContractStub, args [][]byte) ([]byte, error) {
	key := common.IDTypeDID
	kvIterator, err := stub.GetIterator(key+string(common.MinUnicodeValue), key+string(common.MaxUnicodeValue))
	defer kvIterator.Close()
	if err != nil {
		return nil, err
	}
	var array []string
	for kvIterator.Next() {
		key := kvIterator.Key()
		array = append(array, key)
	}
	return json.Marshal(array)
}

func getDocumentPublicKey(document *Document, keyIndex string) (*DocPublicKey, error) {
	if document == nil {
		return nil, errors.New("document empty")
	}
	// passed in did without fragment, use master key
	keyPiece := strings.Split(keyIndex, common.DIDFragmentTag)
	if len(keyPiece) == 1 {
		keyIndex = document.Authentication[0]
	}
	publicKey := &DocPublicKey{}
	for k, v := range document.PublicKey {
		if v.Id == keyIndex {
			*publicKey = document.PublicKey[k]
			break
		}
	}
	if publicKey == nil {
		return nil, errors.New("load master key Error")
	}
	return publicKey, nil
}
